/*

(C) 2009 Mika Ilmaranta <ilmis@nullnet.fi>

License: GPLv2

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "config.h"

static CONFIG defaults;
GLOBAL cfg;

static int eqcmp(char *str, char *pat)
{
	int i;

	i = strlen(pat);
	if(!strncmp(str, pat, i) && str[i] == '=') return 0;
	return 1;
}

int reload_config(char *fn, CONFIG **first, CONFIG **last, GROUPS **firstg, GROUPS **lastg) {
	free_config(first, last, firstg, lastg);
	return(read_config(fn, first, last, firstg, lastg));
}

void free_config(CONFIG **first, CONFIG **last, GROUPS **firstg, GROUPS **lastg) {
	CONFIG *cur, *prev;
	GROUPS *curg, *prevg;
	GROUP_MEMBERS *curgm, *prevgm;

	cur = (*first);
	while(cur) {
		if(cur->name && cur->name != defaults.name)                            free(cur->name);
		if(cur->checkip && cur->checkip != defaults.checkip)                   free(cur->checkip);
		if(cur->eventscript && cur->eventscript != defaults.eventscript)       free(cur->eventscript);
		if(cur->warn_email && cur->warn_email != defaults.warn_email)          free(cur->warn_email);
		if(cur->sourceip && cur->sourceip != defaults.sourceip)                free(cur->sourceip);
		if(cur->device && cur->device != defaults.device)                      free(cur->device);

		prev = cur;
		cur = cur->next;
		free(prev);
	}

	if(defaults.name)                  free(defaults.name);
	if(defaults.checkip)               free(defaults.checkip);
	if(defaults.eventscript)           free(defaults.eventscript);
	if(defaults.warn_email)            free(defaults.warn_email);
	if(defaults.sourceip)              free(defaults.sourceip);
	if(defaults.device)                free(defaults.device);

	(*first) = NULL;
	(*last) = NULL;

	curg = (*firstg);
	while(curg) {
		curgm = curg->fgm;
		while(curgm) {
			free(curgm->name);

			prevgm = curgm;
			curgm = curgm->next;
			free(prevgm);
		}
		curg->fgm = NULL;
		curg->lgm = NULL;

		free(curg->name);
		free(curg->eventscript);
		free(curg->warn_email);

		prevg = curg;
		curg = curg->next;
		free(prevg);
	}

	(*firstg) = NULL;
	(*lastg) = NULL;
}

int read_config(char *fn, CONFIG **first, CONFIG **last, GROUPS **firstg, GROUPS **lastg) {
	static int init_done = 0;

	CONFIG *cur = NULL;
	GROUPS *curg = NULL;
	GROUP_MEMBERS *curgm = NULL;
	FILE *fp;
	char buf[BUFSIZ];
	int mode = 0;
	int error = 0;

	if((fp = fopen(fn, "r")) == 0) {
		syslog(LOG_ERR, "read_config: can't open config file \"%s\"", fn);
		return(-1);
	}

	if(!init_done) { /* handle recursive calls correctly */
		cfg.debug = 8; /* initialize to sane value */
		cfg.reopen_on_enodev = 0; /* assume no reopen on ENODEV */
		defaults.status = 1; /* assume default up state for connections unless user has stated otherwise later in config */

		init_done = 1;
	}

	while(fgets(buf, BUFSIZ, fp)) {
		if(*buf && buf[strlen(buf) - 1] == '\n') buf[strlen(buf) - 1] = 0; /* strip lf */
		if(strchr(buf, '#')) *(strchr(buf, '#')) = 0; /* strip comment */
		while(strchr(buf, '\t')) *(strchr(buf, '\t')) = ' '; /* tab -> space */
		while(*buf == ' ') strcpy(buf, buf + 1); /* strip leading space */
		while(*buf && buf[strlen(buf) - 1] == ' ') buf[strlen(buf) - 1] = 0; /* strip tailing space */

		if(!*buf) continue;

		if (!strncmp(buf, "include ", 8)) {
			read_config(strchr(buf, ' ') + 1, first, last, firstg, lastg);
			continue;
		}

		if(mode) {
			if (!strcmp(buf, "}")) {
				mode=0;
				continue;
			}

			switch(mode) {
			case 1:
				if(!eqcmp(buf, "name"))                       defaults.name                          = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "checkip"))                    defaults.checkip                       = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "eventscript"))                defaults.eventscript                   = strdup(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "max_packet_loss"))            defaults.max_packet_loss               = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "max_successive_pkts_lost"))   defaults.max_successive_pkts_lost      = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "min_packet_loss"))            defaults.min_packet_loss               = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "min_successive_pkts_rcvd"))   defaults.min_successive_pkts_rcvd      = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "interval_ms"))                defaults.interval_ms                   = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "timeout_ms"))                 defaults.timeout_ms                    = atoi(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "warn_email"))                 defaults.warn_email                    = strdup(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "check_arp"))                  defaults.check_arp                     = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "sourceip"))                   defaults.sourceip                      = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "device"))                     defaults.device                        = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "ttl"))                        defaults.ttl                           = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "status"))                     defaults.status                        = atoi(strchr(buf, '=') + 1);

				break;
			case 2:
				if(!cur) {
					syslog(LOG_ERR, "read_config: cur == NULL");
					break;
				}

				if(!eqcmp(buf, "name"))                       cur->name                          = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "checkip"))                    cur->checkip                       = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "eventscript"))                cur->eventscript                   = strdup(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "max_packet_loss"))            cur->max_packet_loss               = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "max_successive_pkts_lost"))   cur->max_successive_pkts_lost      = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "min_packet_loss"))            cur->min_packet_loss               = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "min_successive_pkts_rcvd"))   cur->min_successive_pkts_rcvd      = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "interval_ms"))                cur->interval_ms                   = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "timeout_ms"))                 cur->timeout_ms                    = atoi(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "warn_email"))                 cur->warn_email                    = strdup(strchr(buf, '=') + 1);

				if(!eqcmp(buf, "check_arp"))                  cur->check_arp                     = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "sourceip"))                   cur->sourceip                      = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "device"))                     cur->device                        = strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "ttl"))                        cur->ttl                           = atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "status"))                     cur->status                        = atoi(strchr(buf, '=') + 1);

				break;
			case 3:
				if(!eqcmp(buf, "name"))				curg->name				= strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "eventscript"))			curg->eventscript			= strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "warn_email"))			curg->warn_email			= strdup(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "logic"))			curg->logic				= atoi(strchr(buf, '=') + 1);
				if(!eqcmp(buf, "member-connection")) {
					if((curgm = (GROUP_MEMBERS *)malloc(sizeof(GROUP_MEMBERS))) == NULL) {
						syslog(LOG_ERR, "read_config: can't malloc for group member");
						return(-1);
					}
					curgm->name = strdup(strchr(buf, '=') + 1);
					curgm->cfg_ptr = NULL;

					if(curg->lgm) { /* insert as last */
						curgm->next = NULL;
						curgm->prev = curg->lgm;
						curg->lgm->next = curgm;
						curg->lgm = curgm;
					} else { /* empty member list */
						curgm->next = NULL;
						curgm->prev = NULL;
						curg->fgm = curgm;
						curg->lgm = curgm;
					}
				}
				break;
			default:
				syslog(LOG_ERR, "read_config: switch(mode) default: should never happen mode was %d", mode);
				break;
			}
		}
		else {
			/* global config */
			if(!eqcmp(buf, "debug"))                        cfg.debug                          = atoi(strchr(buf, '=') + 1);
			if(!eqcmp(buf, "reopen_on_enodev"))		cfg.reopen_on_enodev               = atoi(strchr(buf, '=') + 1);

			/* per connection configs */
			if(!strcmp(buf, "defaults {"))
				mode=1;
			if(!strcmp(buf, "connection {")) {
				mode=2;
				if((cur = malloc(sizeof(CONFIG))) == NULL) {
					syslog(LOG_ERR, "read_config: can't malloc for config");
					return(-1);
				}

				if(*last) { /* not first */
					(*last)->next = cur;
					cur->prev = *last;
					cur->next = NULL;
					*last = cur;
				}
				else {
					*first = cur;
					*last = cur;
					cur->prev = NULL;
					cur->next = NULL;
				}
				/* fill in defaults */
				if(defaults.name) {
					cur->name                       = defaults.name;
					cur->checkip                    = defaults.checkip;
					cur->eventscript                = defaults.eventscript;
					cur->max_packet_loss            = defaults.max_packet_loss;
					cur->max_successive_pkts_lost   = defaults.max_successive_pkts_lost;
					cur->min_packet_loss            = defaults.min_packet_loss;
					cur->min_successive_pkts_rcvd   = defaults.min_successive_pkts_rcvd;
					cur->interval_ms                = defaults.interval_ms;
					cur->timeout_ms                 = defaults.timeout_ms;
					cur->warn_email                 = defaults.warn_email;
					cur->check_arp                  = defaults.check_arp;
					cur->sourceip                   = defaults.sourceip;
					cur->device                     = defaults.device;
					cur->ttl                        = defaults.ttl;
					cur->status			= defaults.status;
				}
				else
					syslog(LOG_ERR, "read_config: defaults not set");
			}
			if(!strcmp(buf, "group {")) {
				mode = 3;

				if((curg = (GROUPS *)malloc(sizeof(GROUPS))) == NULL) {
					syslog(LOG_ERR, "read_config: can't malloc for group");
					return(-1);
				}

				/* apply sane defaults for group */
				curg->name            = defaults.name;
				curg->eventscript     = defaults.eventscript;
				curg->warn_email      = defaults.warn_email;
				curg->logic           = 0; /* default group logic or */
				curg->status          = 1; /* default group status is up */

				curg->fgm = NULL;
				curg->lgm = NULL;

				if(*lastg) { /* not first group */
					(*lastg)->next = curg;
					curg->prev = *lastg;
					curg->next = NULL;
					*lastg = curg;
				} else {
					*firstg = curg;
					*lastg = curg;
					curg->prev = NULL;
					curg->next = NULL;
				}
			}
		}
	}

	fclose(fp);

	curg = *firstg;
	while(curg) {
		curgm = curg->fgm;
		while(curgm) {
			int found = 0;

			cur = *first;
			while(cur) {
				if(!strcmp(cur->name, curgm->name)) {
					curgm->cfg_ptr = cur;
					found = 1;
					break;
				}

				cur = cur->next;
			}
			if(!found) {
				syslog(LOG_ERR, "connection group member \"%s\" not found", curgm->name);
				return(-1);
			}
			curgm = curgm->next;
		}
		curg = curg->next;
	}

	/* some parameter sanity checking */
	cur = *first;
	error = 0;
	while(cur) {
		if(cur->max_packet_loss <= cur->min_packet_loss) {
			syslog(LOG_ERR, "connection \"%s\" max_packet_loss <= min_packet_loss. that would cause flip-flop effect", cur->name);
			error++;
		}
		cur = cur->next;
	}
	if(error) return(-1);

	return(0);
}

void dump_config(CONFIG **first, CONFIG **last, GROUPS **firstg, GROUPS **lastg)
{
	CONFIG *cur;
	GROUPS *curg;
	GROUP_MEMBERS *curgm;

	syslog(LOG_INFO,   "cfg.debug                     = \"%d\"", cfg.debug);

	cur = *first;
	while(cur) {
		syslog(LOG_INFO, "cur->name                     = \"%s\"", cur->name);
		syslog(LOG_INFO, "cur->checkip                  = \"%s\"", cur->checkip);
		syslog(LOG_INFO, "cur->eventscript              = \"%s\"", cur->eventscript);

		syslog(LOG_INFO, "cur->max_packet_loss          = \"%d\"", cur->max_packet_loss);
		syslog(LOG_INFO, "cur->max_successive_pkts_lost = \"%d\"", cur->max_successive_pkts_lost);
		syslog(LOG_INFO, "cur->min_packet_loss          = \"%d\"", cur->min_packet_loss);
		syslog(LOG_INFO, "cur->min_successive_pkts_rcvd = \"%d\"", cur->min_successive_pkts_rcvd);
		syslog(LOG_INFO, "cur->interval_ms              = \"%d\"", cur->interval_ms);
		syslog(LOG_INFO, "cur->timeout_ms               = \"%d\"", cur->timeout_ms);

		syslog(LOG_INFO, "cur->warn_email               = \"%s\"", cur->warn_email);

		syslog(LOG_INFO, "cur->check_arp                = \"%d\"", cur->check_arp);
		syslog(LOG_INFO, "cur->sourceip                 = \"%s\"", cur->sourceip);
		syslog(LOG_INFO, "cur->device                   = \"%s\"", cur->device);
		syslog(LOG_INFO, "cur->ttl                      = \"%d\"", cur->ttl);
		syslog(LOG_INFO, "cur->status                   = \"%d\"", cur->status);

		cur = cur->next;
	}

	curg = *firstg;
	while(curg) {
		syslog(LOG_INFO, "curg->name                    = \"%s\"", curg->name);
		syslog(LOG_INFO, "curg->eventscript             = \"%s\"", curg->eventscript);
		syslog(LOG_INFO, "curg->warn_email              = \"%s\"", curg->warn_email);
		syslog(LOG_INFO, "curg->logic                   = \"%s\"", curg->logic == 0 ? "OR" : "AND");

		curgm = curg->fgm;
		while(curgm) {
			syslog(LOG_INFO, "curgm->name                   = \"%s\"", curgm->name);

			curgm = curgm->next;
		}

		curg = curg->next;
	}
}

/* EOF */
